"""ClassDef for dealing with classes mapped with SQLAlchemy."""

from amfast import class_def
from amfast.class_def.as_types import AsNoProxy

# Import from different places depending on which 
# version of SA is being used.
try:
    from sqlalchemy.orm import class_mapper, object_mapper
except ImportError:
    from sqlalchemy.orm.util import class_mapper, object_mapper

# Exception is different depending on which
# version of SA is being used.
UnmappedInstanceError = None
try:
    class_mapper(dict)
except Exception, e:
    UnmappedInstanceError = e.__class__

class SaClassDef(class_def.ClassDef):
    """Defines how objects with a class mapped by SQLAlchemy should be serialized and de-serialized.

    Mapped attributes are considered static.
    Dynamic attributes are ignored.

    The class must be mapped with SQLAlchemy BEFORE calling ClassDefMapper.map_class().
    """
    KEY_ATTR = 'sa_key' # sa instance key
    LAZY_ATTR = 'sa_lazy' # list of lazy attribute names

    # Set to True to always encode
    # KEY_ATTR and LAZY_ATTR as Array objects.
    no_proxy_sa_attrs = True

    def __init__(self, class_, alias=None, static_attrs=None, amf3=None,
        encode_types=None, decode_types=None):
        """Static attributes are inferred from the class mapper,
        so static_attrs needs to be passed only if there are additional
        un-mapped attributes that need to be considered static."""

        try:
            self.mapper = class_mapper(class_)
            self.mapper.compile()
        except UnmappedInstanceError:
            raise class_def.ClassDefError("Class does not have a SA mapper associated with it.")

        if static_attrs is None:
            static_attrs = ()
        self.unmapped_attrs = static_attrs

        self.mapped_attrs = []
        for prop in self.mapper.iterate_properties:
            self.mapped_attrs.append(prop.key)

        # Check for duplicates
        for attr in self.mapped_attrs:
            if attr in self.unmapped_attrs:
                raise class_def.ClassDefError("Mapped attributes cannot be listed in the static_attrs argument.")

        combined_attrs = [self.KEY_ATTR, self.LAZY_ATTR]
        combined_attrs.extend(self.mapped_attrs)
        combined_attrs.extend(self.unmapped_attrs)

        class_def.ClassDef.__init__(self, class_, alias=alias,
            static_attrs=combined_attrs, amf3=amf3, encode_types=encode_types,
            decode_types=decode_types)

    def getStaticAttrVals(self, obj):
        # Set key and lazy
        lazy_attrs = []
        
        if self.__class__.no_proxy_sa_attrs is True:
            vals = [AsNoProxy(self.mapper.primary_key_from_instance(obj)), AsNoProxy(lazy_attrs)]
        else:
            vals = [self.mapper.primary_key_from_instance(obj), lazy_attrs]

        # Set mapped values
        attr_count = len(self.mapped_attrs)
        for i in xrange(0, attr_count):
            attr = self.mapped_attrs[i]
 
            # Look at __dict__ directly,
            # otherwise SA will touch the attr.
            if attr in obj.__dict__:
                vals.append(getattr(obj, attr))
            else:
                # This attr is lazy
                vals.append(None)
                lazy_attrs.append(attr)

        # Set un-mapped values
        vals.extend([getattr(obj, attr, None) for attr in self.unmapped_attrs])

        return vals

    def getInstance(self):
        return self.mapper.class_manager.new_instance()

    def applyAttrVals(self, obj, vals):
        # Delete lazy-loaded attrs from vals
        if self.LAZY_ATTR in vals:
            lazy_attrs = vals[self.LAZY_ATTR]
            if lazy_attrs is not None:
                for lazy_attr in lazy_attrs:
                    if lazy_attr in vals:
                        del vals[lazy_attr]
            del vals[self.LAZY_ATTR]

        class_def.ClassDef.applyAttrVals(self, obj, vals)



"""Generates Actionscript code from ClassDef objects."""
import os

class CodeGenerator(object):
    """Generates Actionscript source code from class defs.

    attributes:
    ============
     * indent - string, indention string to use. Default = '\t'
    """

    def __init__(self, indent='\t'):
        self.indent = indent

    def generateFilesFromMapper(self, class_mapper, dir='.', use_accessors=False,
        packaged=False, constructor=False, bindable=False,
        extends=None, implements=None):
        """Generates an Actionscript class files.

        arguments:
        ===========
         * class_mapper - amfast.class_def.ClassDefMapper, ClassDefMapper being used to generate the action script classes.
         * dir - string, directory to store generated files in. Default = '.'
         * use_accessors - bool, If True create getter and setter methods for attributes. Default = False
         * packaged - bool, If True wrap created class in a package (for AS3). Default = False
         * constructor - bool, If True create a constructor method. Default = False
         * bindable - bool, If True, add the 'Bindable' metadata tag. Default = False
         * extends - string, name of the class that this class inherits from. Default = None
         * implements - list or tuple, a list of interface names that this class implments. Default = None
        """
        for class_def in class_mapper:
            if class_def._built_in is True:
                continue

            self.generateClassFile(class_def, dir, use_accessors,
                packaged, constructor, bindable, extends, implements)

    def generateClassFile(self, class_def, dir='.', use_accessors=False,
        packaged=False, constructor=False, bindable=False,
        extends=None, implements=None):
        """Generates an Actionscript class source file.

        arguments:
        ===========
         * class_def - amfast.class_def.ClassDef, ClassDef being used to generate the action script class.
         * dir - string, directory to store generated files in. Default = '.'
         * use_accessors - bool, If True create getter and setter methods for attributes. Default = False
         * packaged - bool, If True wrap created class in a package (for AS3). Default = False
         * constructor - bool, If True create a constructor method. Default = False
         * bindable - bool, If True, add the 'Bindable' metadata tag. Default = False
         * extends - string, name of the class that this class inherits from. Default = None
         * implements - list or tuple, a list of interface names that this class implments. Default = None
        """
        package = class_def.alias.split('.')
        class_ = package.pop()

        # Create all output directories
        path = dir
        for i in range(0, len(package)):
            path = os.path.join(path, package[i])
            if os.path.exists(path) is not True:
                os.mkdir(path)

        out = open(os.path.join(path, '%s.as' % class_), 'w')
        out.write(self.generateClassStr(class_def, use_accessors, packaged,
            constructor, bindable, extends, implements))
        out.close()

    def generateClassStr(self, class_def, use_accessors=False,
        packaged=False, constructor=False, bindable=False,
        extends=None, implements=None):
        """Generates an Actionscript class source string.

        arguments:
        ===========
         * class_def - amfast.class_def.ClassDef, ClassDef being used to generate the action script class.
         * use_accessors - bool, If True create getter and setter methods for attributes. Default = False
         * packaged - bool, If True wrap created class in a package (for AS3). Default = False
         * constructor - bool, If True create a constructor method. Default = False
         * bindable - bool, If True, add the 'Bindable' metadata tag. Default = False
         * extends - string, name of the class that this class inherits from. Default = None
         * implements - list or tuple, a list of interface names that this class implments. Default = None
        """
        class_str = []

        package = class_def.alias.split('.')
        class_ = package.pop()
        package = '.'.join(package)
        indent = ''

        if packaged is True:
           class_str.append('package %s' % package)
           class_str.append('{')
           indent = self.indent

        if bindable is True:
            class_str.append(indent + '[Bindable]')

        class_str.append(indent + "[RemoteClass(alias='%s')]" % class_def.alias)

        class_def_str = indent
        if packaged is True:
            class_def_str += 'public '
        
        if hasattr(class_def, "DYNAMIC_CLASS_DEF") is True:
            class_def_str += 'dynamic '

        if hasattr(class_def, "EXTERNIZEABLE_CLASS_DEF") is True:
            imp = ['IExternalizable']
            if implements is not None:
                imp.extend(implements)
            implements = imp

        class_def_str += 'class %s' % class_
        if extends is not None:
            class_def_str += ' extends %s' % extends

        if implements is not None and len(implements) > 0:
            class_def_str += ' implements %s' % ', '.join(implements)

        class_str.append(class_def_str);
        class_str.append(indent + '{')
        for attr in class_def.static_attrs:
            if use_accessors is True:
                class_str.append(self.generateAccessor(attr, indent + self.indent))
            else:
                class_str.append(indent + self.indent + 'public var %s:Object;' % attr)

        if constructor is True:
            class_str.append('\n' + indent + self.indent + 'public function %s():void' % class_)
            class_str.append(indent + self.indent + '{')
            if extends is not None:
                class_str.append(indent + self.indent + self.indent + 'super();')
            class_str.append(indent + self.indent + '}')

        class_str.append(indent + '}')
        if packaged is True:
            class_str.append('}')
        return '\n'.join(class_str)

    def generateAccessor(self, attr, indent=''):
        """Generates an Actionscript getter and setter source string.

        arguments:
        ===========
         * attr - string, the attribute name to generate code for.
         * indent - string, indent to add to generated code. Default = ''.
        """
        attr_str = []
        attr_str.append('\n' + indent + "private var _%s:Object;" % attr)

        attr_str.append(indent + "public function get %s():Object" % attr)
        attr_str.append(indent + "{")
        attr_str.append(indent + self.indent + "return _%s;" % attr)
        attr_str.append(indent + "}")

        attr_str.append('\n' + indent + "public function set %s(value:Object):void" % attr)
        attr_str.append(indent + "{")
        attr_str.append(indent + self.indent + " _%s = value;" % attr)
        attr_str.append(indent + "}")

        return '\n'.join(attr_str);



import time

import memcache_manager
from connection import Connection, ConnectionError
from connection_manager import ConnectionManager, NotConnectedError, SessionAttrError

class MemcacheConnectionManager(ConnectionManager, memcache_manager.MemcacheManager):
    """Manages connections stored by Memcache."""

    CONNECTIONS_ATTR = '_connection_ids'
    CHANNELS_ATTR = '_channels'
    ATTRIBUTES = ('connection_info', 'connected', 'last_active',
        'last_polled', 'authenticated', 'flex_user', 'session', 'notify_func_id')

    def __init__(self, connection_class=Connection, connection_params=None,
        mc_servers=['127.0.0.1:11211'], mc_debug=0):
        ConnectionManager.__init__(self, connection_class=connection_class,
            connection_params=connection_params)

        self.mc = self.createMcClient(mc_servers, mc_debug)
        self._lock = memcache_manager.MemcacheMutex(self.mc)

    def reset(self):
        self._lock.releaseAll()

        lock_name = self.getLockName('connection_reset')
        self._lock.acquire(lock_name)
        try:
            connection_ids = self.mc.get(self.CONNECTIONS_ATTR)
            if connection_ids is not None:
                for connection_id in connection_ids:
                    keys = [self.getKeyName(connection_id, attr) for attr in self.ATTRIBUTES]
                    self.mc.delete_multi(keys)

                self.mc.set(self.CONNECTIONS_ATTR, [])
                self.mc.set(self.CHANNELS_ATTR, {})
        finally:
            self._lock.release(lock_name)

    def incrementChannelCount(self, channel_name):
        lock_name = self.getLockName(self.CHANNELS_ATTR)
        self._lock.acquire(lock_name)
        try:
            channels = self.mc.get(self.CHANNELS_ATTR)
            if channels is None:
                channels = {}
            if channel_name in channels:
                channels[channel_name] += 1
            else:
                channels[channel_name] = 1
            
            self.mc.set(self.CHANNELS_ATTR, channels)
        finally:
            self._lock.release(lock_name)

    def decrementChannelCount(self, channel_name):
        lock_name = self.getLockName(self.CHANNELS_ATTR)
        self._lock.acquire(lock_name)
        try:
            channels = self.mc.get(self.CHANNELS_ATTR)
            if channels is None:
                channels = {}
            if channel_name in channels:
                channels[channel_name] -= 1
                self.mc.set(self.CHANNELS_ATTR, channels)
        finally:
            self._lock.release(lock_name)

    def getConnectionCount(self, channel_name):
        channels = self.mc.get(self.CHANNELS_ATTR)
        if channels is None:
            return 0

        if channel_name in channels:
            return channels[channel_name]
        else:
            return 0

    def checkMultiSetResults(self, results):
        if len(results) > 0:
           msg = 'The following parameters were not set: ' + ', '.join(results)
           raise ConnectionError(msg)

    def loadConnection(self, connection_id):
        connection_info = self.mc.get(self.getKeyName(connection_id, 'connection_info'))
        
        if connection_info is None:
            raise NotConnectedError("Connection '%s' is not connected." % connection_id)

        return self.connection_class(self, connection_info['channel_name'],
            connection_id, connection_info['timeout'])

    def initConnection(self, connection, channel):
        params = {
            'connected': True,
            'last_active': time.time() * 1000,
            'last_polled': 0.0,
            'authenticated': False,
            'session': {}
        }

        connection_info = {
            'channel_name': connection.channel_name,
            'timeout': connection.timeout
        }

        cache_params = {}
        for key, val in params.iteritems():
            cache_params[self.getKeyName(connection.id, key)] = val
        cache_params[self.getKeyName(connection.id, 'connection_info')] = connection_info

        self.checkMultiSetResults(self.mc.set_multi(cache_params))

        lock_name = self.getLockName(self.CONNECTIONS_ATTR)
        self._lock.acquire(lock_name)
        try:
            connection_ids = self.mc.get(self.CONNECTIONS_ATTR)
            if connection_ids is None:
                connection_ids = []
            connection_ids.append(connection.id)
            self.mc.set(self.CONNECTIONS_ATTR, connection_ids)
        finally:
            self._lock.release(lock_name)

        self.incrementChannelCount(connection.channel_name)

    def iterConnectionIds(self):
        connection_ids = self.mc.get(self.CONNECTIONS_ATTR)
        if connection_ids != None:
          return connection_ids.__iter__()
   else:
          return None 

    # --- proxies for connection properties --- #

    def getConnected(self, connection):
        return self.mc.get(self.getKeyName(connection.id, 'connected'))

    def getLastActive(self, connection):
        return self.mc.get(self.getKeyName(connection.id, 'last_active'))

    def getLastPolled(self, connection):
        return self.mc.get(self.getKeyName(connection.id, 'last_polled'))

    def getAuthenticated(self, connection):
        return self.mc.get(self.getKeyName(connection.id, 'authenticated'))

    def getFlexUser(self, connection):
        return self.mc.get(self.getKeyName(connection.id, 'flex_user'))

    def getNotifyFunc(self, connection):
        notify_func_id = self.mc.get(self.getKeyName(connection.id, 'notify_func_id'))

        if notify_func_id is None:
            return None
        else:
            return connection._getNotifyFuncById(connection._notify_func_id)

    # --- proxies for connection methods --- #

    def deleteConnection(self, connection):
        lock_name = self.getLockName(self.CONNECTIONS_ATTR)
        self._lock.acquire(lock_name)
        try:
            connection_ids = self.mc.get(self.CONNECTIONS_ATTR)
            for i, connection_id in enumerate(connection_ids):
                if connection_id == connection.id:
                    connection_ids.pop(i)
                    break
            self.mc.set(self.CONNECTIONS_ATTR, connection_ids)
        finally:
            self._lock.release(lock_name)

        keys = [self.getKeyName(connection.id, attr) for attr in self.ATTRIBUTES]
        self.mc.delete_multi(keys)

        self.decrementChannelCount(connection.channel_name)

        ConnectionManager.deleteConnection(self, connection)

    def connectConnection(self, connection):
        self.mc.set(self.getKeyName(connection.id, 'connected'), True)

    def disconnectConnection(self, connection):
        self.mc.set(self.getKeyName(connection.id, 'connected'), False)

    def touchConnection(self, connection):
        self.mc.set(self.getKeyName(connection.id, 'last_active'), time.time() * 1000)

    def touchPolled(self, connection):
        self.mc.set(self.getKeyName(connection.id, 'last_polled'), time.time() * 1000)

    def authenticateConnection(self, connection, user):
        params = {
            self.getKeyName(connection.id, 'authenticated'): True,
            self.getKeyName(connection.id, 'flex_user'): user
        }

        self.checkMultiSetResults(self.mc.set_multi(params))

    def unAuthenticateConnection(self, connection):
        self.mc.set(self.getKeyName(connection.id, 'authenticated'), False)
        self.mc.delete(self.getKeyName(connection.id, 'flex_user'))

    def setNotifyFunc(self, connection, func):
        self.mc.set(self.getKeyName(connection.id, 'notify_func_id'),
            connection._setNotifyFunc(func))

    def unSetNotifyFunc(self, connection):
        self.mc.delete(self.getKeyName(connection.id, 'notify_func_id'))

    def getConnectionSessionAttr(self, connection, name):
        session = self.mc.get(self.getKeyName(connection.id, 'session'))
        try:
            return session[name]
        except KeyError:
            raise SessionAttrError("Attribute '%s' not found." % name)

    def setConnectionSessionAttr(self, connection, name, val):
        key = self.getKeyName(connection.id, 'session')

        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            session = self.mc.get(key)
            session[name] = val
            self.mc.set(key, session)
        finally:
            self._lock.release(lock_name)

    def delConnectionSessionAttr(self, connection, name):
        key = self.getKeyName(connection.id, 'session')

        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            session = self.mc.get(key)
            try:
                del session[name]
                self.mc.set(key, session)
            except KeyError:
                pass
        finally:
            self._lock.release(lock_name)



import time

import tornado.web
from tornado.ioloop import IOLoop, PeriodicCallback

import amfast
from amfast.class_def.as_types import AsNoProxy
from channel import ChannelSet, HttpChannel
from endpoint import AmfEndpoint
import flex_messages as messaging
import connection_manager as cm

class TornadoChannelSet(ChannelSet):
    def scheduleClean(self):
        cleaner = PeriodicCallback(self.clean, self.clean_freq * 1000,
            IOLoop.instance())
        cleaner.start()

    def clean(self):
        if amfast.log_debug is True:
            amfast.logger.debug("Cleaning channel.")

        current_time = time.time()
        iter = self.connection_manager.iterConnectionIds()
        def _clean():
            try:
                connection_id = iter.next()
                def _callback():
                    self.cleanConnection(connection_id, current_time)
                IOLoop.instance().add_callback(_callback)
                IOLoop.instance().add_callback(_clean)
            except StopIteration:
                pass
        _clean()

    def notifyConnections(self, topic, sub_topic):
        for connection_id in self.subscription_manager.iterSubscribers(topic, sub_topic):
            self.notifyConnection(connection_id)

    def notifyConnection(self, connection_id):
        def _notify():
            self._notifyConnection(connection_id)
        IOLoop.instance().add_callback(_notify)

    def _notifyConnection(self, connection_id):
        try:
            connection = self.connection_manager.getConnection(connection_id, False)
        except cm.NotConnectedError:
            return

        if connection.notify_func is not None:
            IOLoop.instance().add_callback(connection.notify_func)

class CallbackChain(object):
    """Chain together Tornado callbacks.

    When a callback completes, the next in line is called with
    the output of the previous, until the chain is completed."""

    def __init__(self):
        self.callbacks = []

    def addCallback(self, function, args=[], kwargs={}):
        self.callbacks.append({
            'function': function,
            'args': args,
            'kwargs': kwargs
        })

    def execute(self, arg=None):
        """Execute the callback chain.

        arguments:
        ============
         * arg - object, Argument to pass to 1st callback. Default = None
        """
        callback_cnt = len(self.callbacks)
        if callback_cnt < 1:
            return

        callback = self.callbacks.pop(0)

        def _execute():
            result = callback['function'](arg, *callback['args'],
                **callback['kwargs'])
            if callback_cnt > 1:
                self.execute(result)

        IOLoop.instance().add_callback(_execute)

class TornadoChannel(HttpChannel):

    # This attribute is added to packets
    # that are waiting for a long-poll
    # to receive a message.
    MSG_NOT_COMPLETE = '_msg_not_complete'

    # This attribute is added to store
    # Tornado's request handler on the packet,
    # so that it can be available to targets.
    TORNADO_REQUEST = '_tornado_request'

    def __init__(self, *args, **kwargs):

        HttpChannel.__init__(self, *args, **kwargs)

        class requestHandler(tornado.web.RequestHandler):
            """The RequestHandler class for this Channel."""

            @tornado.web.asynchronous
            def post(inner_self):
                """Process an incoming request."""
		self.processRequest(inner_self)

        self.request_handler = requestHandler

    def processRequest(self, request_handler):
        request_handler.set_header('Content-Type', self.CONTENT_TYPE)

        call_chain = CallbackChain()
        call_chain.addCallback(self.decode)
        call_chain.addCallback(self.invoke)
        call_chain.addCallback(self.checkComplete, (request_handler,))
        call_chain.execute(request_handler)

    def decode(self, request_handler):
        """Overridden to add Tornado's request object onto the packet."""
        decoded = HttpChannel.decode(self, request_handler.request.body)
        setattr(decoded, self.TORNADO_REQUEST, request_handler)
        return decoded

    def checkComplete(self, response, request_handler):
        """Checks to determine if the response message is ready
        to be returned to the client, and finishes the request if ready.
        """

        if hasattr(response, self.MSG_NOT_COMPLETE):
            # long-poll operation.
            # response is waiting for a message to be published.
            return

        if request_handler._finished is True:
            # Someone else already finished the request.
            return

        if request_handler.request.connection.stream.closed():
            # Client is not connected.
            return

        # Message is complete, encode and return
        request_handler.finish(self.encode(response))

    def setupPollRequest(self, packet):
        """Setup a request for a long-poll operation."""

        # Set flag so self.checkComplete
        # does not finish the message.
        setattr(packet.response, self.MSG_NOT_COMPLETE, True)

        return getattr(packet, self.TORNADO_REQUEST)

    def finishPoll(self, request, packet, message, messages):
        """Finish a request that has been waiting for messages."""

        if isinstance(packet.channel.endpoint, AmfEndpoint):
            # Make sure messages are not encoded as an ArrayCollection
            messages = AsNoProxy(messages)
        message.response_msg.body.body = messages

        if hasattr(packet.response, self.MSG_NOT_COMPLETE):
            delattr(packet.response, self.MSG_NOT_COMPLETE)
        self.checkComplete(packet.response, request)

    def _waitForMessage(self, packet, message, connection):
        """Overridden to be non-blocking."""

        request = self.setupPollRequest(packet)

        def _notify():
            # This function gets called when a message is published,
            # or wait_interval is reached.
            connection.unSetNotifyFunc()

            # Get messages and add them
            # to the response message
            messages = self.channel_set.subscription_manager.pollConnection(connection)
            self.finishPoll(request, packet, message, messages)

        # Setup timeout
        if self.wait_interval > -1:
            timeout_call = IOLoop.instance().add_timeout(
                time.time() + float(self.wait_interval) / 1000, _notify)

        def _notifyTimeout():
            # Notifies, plus cancels the timeout
            if timeout_call is not None:
                IOLoop.instance().remove_timeout(timeout_call)
            _notify()
        connection.setNotifyFunc(_notifyTimeout)

        # Remove notify function if client drops connection.
        request.request.connection.stream.set_close_callback(connection.unSetNotifyFunc)
 
        return ()

    def _pollForMessage(self, packet, message, connection):
        """Overridden to be non-blocking."""

        request = self.setupPollRequest(packet)

        # Polls for messages every self.poll_interval
        poller = PeriodicCallback(None, float(self.poll_interval) / 1000, IOLoop.instance())

        def _timeout():
            # Executed when timeout is reached.
            poller.stop()
            messages = self.channel_set.subscription_manager.pollConnection(connection)
            self.finishPoll(request, packet, message, messages)

        if self.wait_interval > -1:
            timeout_call = IOLoop.instance().add_timeout(
                time.time() + float(self.wait_interval) / 1000, _timeout)
        else:
            timeout_call = None

        # Timeout if client drops connection.
        request.request.connection.stream.set_close_callback(_timeout)

        def _poll():
            messages = self.channel_set.subscription_manager.pollConnection(connection)
            if len(messages) > 0:
                poller.stop()
                if timeout_call is not None:
                    # Disable time out callback
                    IOLoop.instance().remove_timeout(timeout_call)
                self.finishPoll(request, packet, message, messages)

        poller.callback = _poll
	poller.start()

        return ()

class StreamingTornadoChannel(TornadoChannel):
    """Handles streaming http connections."""

    def __init__(self, name, max_connections=-1, endpoint=None,
        timeout=1200, wait_interval=0, heart_interval=5):
        TornadoChannel.__init__(self, name, max_connections, endpoint,
            timeout, wait_interval)

        self.heart_interval = heart_interval
	
    def processRequest(self, request_handler):
        if request_handler.request.headers['Content-Type'] == self.CONTENT_TYPE:
	    # Regular AMF message
	    return TornadoChannel.processRequest(self, request_handler)

        request_handler.set_header('Content-Type', self.CONTENT_TYPE)
        msg = messaging.StreamingMessage()
	msg.parseArgs(request_handler.request.arguments)

	if msg.operation == msg.OPEN_COMMAND:
	    def _open():
	        self.startStream(msg, request_handler)
	    IOLoop.instance().add_callback(request_handler.async_callback(_open))
	elif msg.operation == msg.CLOSE_COMMAND:
            pass

    def startStream(self, msg, request_handler):
        """Get this stream rolling!"""

        connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))

        if self.channel_set.notify_connections is True:
            poller = None
        else:
            # Call _notify multiple times if polling.
            poller = PeriodicCallback(None, float(self.poll_interval) / 1000, IOLoop.instance())

        # Handle new message.
	def _notify():
            if connection.connected is False:
                if poller is not None:
                    poller.stop()

	        connection.unSetNotifyFunc()
		if request_handler.request.connection.stream.closed() is False:
		    msg = messaging.StreamingMessage.getDisconnectMsg()
		    self.sendMsgs((msg,), request_handler)
		    request_handler.finish()
		return
		    
	    msgs = self.channel_set.subscription_manager.pollConnection(connection)
            if len(msgs) > 0:
	        self.sendMsgs(msgs, request_handler)
	connection.setNotifyFunc(_notify)

        if poller is not None:
           poller.callback = _notify
           poller.start() 

        # Handle dropped connection.
	def _connectionLost():
            if poller is not None:
                poller.stop()
	    self.channel_set.disconnect(connection)
	    connection.unSetNotifyFunc()   
        request_handler.request.connection.stream.set_close_callback(_connectionLost)

	# Send acknowledge message
	response = msg.acknowledge()
	response.body = connection.id
	self.sendMsgs((response,), request_handler)

        request_handler.write(chr(messaging.StreamingMessage.NULL_BYTE) * self.KICKSTART_BYTES)
        request_handler.flush()

	self.startBeat(connection, request_handler)

    def sendMsgs(self, msgs, request_handler):
        """Send messages to the client."""
	for msg in msgs:
	    request_handler.write(messaging.StreamingMessage.prepareMsg(msg, self.endpoint))
	request_handler.flush()

    def startBeat(self, connection, request_handler):
        beater = PeriodicCallback(None, self.heart_interval, IOLoop.instance())
        def _beat():
            if connection.connected is False:
	        beater.stop()
	    else:
		request_handler.write(chr(messaging.StreamingMessage.NULL_BYTE))
		request_handler.flush()
	beater.callback = _beat
	beater.start()



import os
import sys
import time
import unittest

from amfast.remoting.connection_manager import NotConnectedError, \
    SessionAttrError, MemoryConnectionManager

class ConnectionTestCase(unittest.TestCase):

    class TestChannel(object):
        def __init__(self):
            self.name = 'test'

    def setUp(self):
        self.manager.reset()
        self.channel = self.TestChannel()

    def _testConnectionProps(self, connection):
        """Checks newly inited props."""
        self.assertEquals(36, len(connection.id))
        self.assertEquals(self.channel.name, connection.channel_name)
        self.assertTrue(connection.connected)
        self.assertFalse(connection.authenticated)
        self.assertEquals(None, connection.flex_user)
        self.assertTrue(connection.last_active > 0)
        self.assertTrue(time.time() * 1000 > connection.last_active)

    def _testCompare(self, connection, connection_2):
        self.assertEquals(connection.id, connection_2.id)
        self.assertEquals(connection.channel_name, connection_2.channel_name)
        self.assertEquals(connection.timeout, connection_2.timeout)
        self.assertEquals(connection.connected, connection_2.connected)
        self.assertEquals(connection.authenticated, connection_2.authenticated)
        self.assertEquals(connection.flex_user, connection_2.flex_user)
        self.assertEquals(connection.last_active, connection_2.last_active)
        self.assertEquals(connection.last_polled, connection_2.last_polled)

    def testCreateConnection(self):
        connection = self.manager.createConnection(self.channel)
        self._testConnectionProps(connection)

    def testGetConnection(self):
        connection = self.manager.createConnection(self.channel)
        last_active = connection.last_active

        new_connection = self.manager.getConnection(connection.id, touch=True)
        self._testConnectionProps(new_connection)
        self.assertTrue(last_active < new_connection.last_active)
        last_active = new_connection.last_active

        new_connection = self.manager.getConnection(connection.id, touch=False)
        self.assertEquals(last_active, new_connection.last_active)

        self._testCompare(connection, new_connection)

    def testGetConnectionRaisesNotConnectedError(self):
        self.assertRaises(NotConnectedError, self.manager.getConnection, 'not_connected')

    def testDeleteConnection(self):
        connection = self.manager.createConnection(self.channel)
        self.manager.deleteConnection(connection)
        self.assertFalse(connection.connected)
        self.assertRaises(NotConnectedError, self.manager.getConnection, connection.id)

    def testConnectConnection(self):
        connection = self.manager.createConnection(self.channel)
        connection.disconnect()
        self.assertFalse(connection.connected)
        connection.connect()
        self.assertTrue(connection.connected)

    def testTouch(self):
        connection = self.manager.createConnection(self.channel)
        last_active = connection.last_active
        connection.touch()
        self.assertTrue(connection.last_active > last_active)

    def testTouchPoll(self):
        connection = self.manager.createConnection(self.channel)
        last_polled = connection.last_polled
        connection.touchPolled()
        self.assertTrue(connection.last_polled > last_polled)

    def testAuthenticate(self):
        connection = self.manager.createConnection(self.channel)
        user = 'tester'
        connection.authenticate(user)
        self.assertTrue(connection.authenticated)
        self.assertEquals(user, connection.flex_user)

        connection.unAuthenticate()
        self.assertFalse(connection.authenticated)
        self.assertEquals(None, connection.flex_user)

    def testNotifyFunc(self):
        def notify():
            return True

        connection = self.manager.createConnection(self.channel)
        connection.setNotifyFunc(notify)
        self.assertTrue(connection.notify_func())
        connection.unSetNotifyFunc()
        self.assertEquals(None, connection.notify_func)

    def testSessionAttrs(self):
        connection = self.manager.createConnection(self.channel)
        key_1 = 'key_1'
        key_2 = 'key_2'
        val_1 = 'val_1'
        val_2 = 'val_2'

        connection.setSessionAttr(key_1, val_1)
        connection.setSessionAttr(key_2, val_2)
        self.assertEquals(val_1, connection.getSessionAttr(key_1))
        self.assertEquals(val_2, connection.getSessionAttr(key_2))

        connection = self.manager.getConnection(connection.id)
        self.assertEquals(val_1, connection.getSessionAttr(key_1))
        self.assertEquals(val_2, connection.getSessionAttr(key_2))

        connection.delSessionAttr(key_1)
        self.assertEquals(val_2, connection.getSessionAttr(key_2))

        connection = self.manager.getConnection(connection.id) 
        self.assertRaises(SessionAttrError, connection.getSessionAttr, key_1)

    def testReset(self):
        connection = self.manager.createConnection(self.channel)
        self.manager.reset()
        self.assertRaises(NotConnectedError, self.manager.getConnection, connection.id)

    def testChannelCount(self):
        count = 5
        ids = []
        for i in xrange(count):
            connection = self.manager.createConnection(self.channel)
            ids.append(connection.id)

        self.assertEquals(count, self.manager.getConnectionCount(self.channel.name))

        for connection_id in ids:
            connection = self.manager.getConnection(connection_id)
            connection.delete()

        self.assertEquals(0, self.manager.getConnectionCount(self.channel.name))

class MemoryTestCase(ConnectionTestCase):

    def setUp(self):
        self.manager = MemoryConnectionManager()
        ConnectionTestCase.setUp(self)

class GaeTestCase(ConnectionTestCase):

    def setUp(self):
        from amfast.remoting.gae_connection_manager import GaeConnectionManager
        self.manager = GaeConnectionManager()
        ConnectionTestCase.setUp(self)

    def testNotifyFunc(self):
        pass

class MemcacheTestCase(ConnectionTestCase):

    def setUp(self):
        from amfast.remoting.memcache_connection_manager import MemcacheConnectionManager
        self.manager = MemcacheConnectionManager()
        ConnectionTestCase.setUp(self)

    def testNotifyFunc(self):
        pass

class SaTestCase(ConnectionTestCase):

    def setUp(self):
        import sqlalchemy as sa
        from amfast.remoting.sa_connection_manager import SaConnectionManager

        engine = sa.create_engine('sqlite:///sa_test_case.db', echo=False)
        metadata = sa.MetaData()

        self.manager = SaConnectionManager(engine, metadata)
        self.manager.createTables()
        ConnectionTestCase.setUp(self)

def suite():
    tests = [
        unittest.TestLoader().loadTestsFromTestCase(MemoryTestCase)
    ]

    print "\n---- Optional Connection Tests ----"
    try:
        from amfast.remoting.gae_connection_manager import GaeConnectionManager
    except Exception:
        # Skip if we're not in Gae environment.
        print "Skipping GAE test."
    else:
        print "Running GAE test."
        tests.append(unittest.TestLoader().loadTestsFromTestCase(GaeTestCase))

    try:
        import sqlalchemy as sa
        from amfast.remoting.sa_connection_manager import SaConnectionManager
    except Exception:
        # Skip if SQLAlchemy is not installed.
        print "Skipping SA test."
    else:
        print "Running SA test."
        tests.append(unittest.TestLoader().loadTestsFromTestCase(SaTestCase))

    try:
        from amfast.remoting.memcache_connection_manager import MemcacheConnectionManager

        # Check connection
        manager = MemcacheConnectionManager()
        if manager.mc.set("test", True) is not True:
            print "Memcache set failed."
            raise Error("Memcache connection failed.")
        if manager.mc.get("test") != True:
            print "Memcache get failed."
            raise Error("Memcache connection failed.")
    except Exception:
        # Skip if memcache support is not installed.
        print "Skipping Memcache test."
    else:
        print "Running Memcache test."
        tests.append(unittest.TestLoader().loadTestsFromTestCase(MemcacheTestCase))

    print "--------"

    return unittest.TestSuite(tests)

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())   



from django import http

import amfast
from channel import HttpChannel, ChannelError

class DjangoChannel(HttpChannel):
    """A channel that works with Django."""

    # Attribute that holds Django's
    # request object, so that it can 
    # be accessed from a target.
    DJANGO_REQUEST = '_django_request'

    def __call__(self, http_request):
        if http_request.method != 'POST':
            return http.HttpResponseNotAllowed(['POST'])

        try:
            request_packet = self.decode(http_request.raw_post_data)
            setattr(request_packet, self.DJANGO_REQUEST, http_request)
        except amfast.AmFastError, exc:
            return http.HttpResponseBadRequest(mimetype='text/plain', content=self.getBadEncodingMsg())
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return http.HttpResponseServerError(mimetype='text/plain', content=self.getBadServerMsg())

        try:
            response_packet = self.invoke(request_packet)
            raw_response = self.encode(response_packet)

            http_response = http.HttpResponse(mimetype=self.CONTENT_TYPE)
            http_response['Content-Length'] = str(len(raw_response))
            http_response.write(raw_response)
            return http_response
        except amfast.AmFastError, exc:
            return http.HttpResponseServerError(mimetype='text/plain', content=self.getBadServerMsg())
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return http.HttpResponseServerError(mimetype='text/plain', content=self.getBadServerMsg())

# TODO: StreamingDjangoChannel!!
